import os
import sys

from setuptools import Command
from setuptools.command.test import test as TestCommand

name = 'pysorter'
base_dir = os.path.dirname(__file__)

def version_path():
    return os.path.join(base_dir, name, 'version.txt')

def read_version():
    return open(version_path()).read().strip()

def write_version(string):
    f = open(version_path(), 'w')
    f.write(string)
    f.write('\n')

# --------------------------------------------------------------------------
#  Custom setup commands
# --------------------------------------------------------------------------
class CoverageCommand(Command):
    description = "generates a coverage report"
    user_options = []

    def initialize_options(self):
        self.cwd = None

    def finalize_options(self):
        self.cwd = os.getcwd()

    def run(self):
        assert os.getcwd() == self.cwd, 'Must be in package root: %s' % self.cwd
        os.system("py.test --cov-report 'html:{}/coverage_report' --cov=pysorter".format(os.getcwd()))

class BumpVersionCommand(Command):
    description = "bumps the versions minor "
    user_options = [
        ('field=', None, 'which field should be bumped: major, minor or patch?')
    ]

    def initialize_options(self):
        self.cwd = None
        self.field = 'patch'
        self.field_index = {'major':0, 'minor':1, 'patch':2}

    def finalize_options(self):
        self.cwd = os.getcwd()
        assert self.field in ('major', 'minor', 'patch'), 'Invalid field!'


    def run(self):
        assert os.getcwd() == self.cwd, 'Must be in package root: %s' % self.cwd
        current = read_version()
        version = [int(_) for _ in current.split('.')]
        version[self.field_index[self.field]] += 1
        for i in range(self.field_index[self.field] + 1, len(self.field_index)):
            version[i] = 0


        version = '.'.join([str(_) for _ in version])
        write_version(version)
        print("version bumped {} --> {}".format(current, version))
    

class PyTestCommand(TestCommand):
    """Setup the py.test test runner."""

    def finalize_options(self):
        """Set options for the command line."""
        TestCommand.finalize_options(self)
        self.test_args = []# ['--cov', name]
        self.test_suite = True

    def run_tests(self):
        """Execute the test runner command."""
        # Import here, because outside the required eggs aren't loaded yet
        import pytest
        sys.exit(pytest.main(self.test_args))



"""
    Contains the core logic of the sorting implementation
"""
from __future__ import print_function

import logging
import os

from .. import filesystem as fs
from ..rules.basesortrule import UnhandledPathException

log = logging.getLogger(__name__)


class PySorter(object):
    def __init__(self,
                 source_dir,
                 sort_rule,
                 unknown_suffix=None,
                 directories_dest=None,
                 no_process=None,
                 other_files=None,
                 dest_dir=None,

                 do_recurse=False,
                 do_remove_empty_dirs=False,
                 do_process_dirs=False):
        """
        Construct a new instance of PySorter for organizing some directory
        using certain parameters

        Parameters
        -----------
        source_dir : string
           path to the directory that must be organized

        sort_rule: RuleSet
            ruleset used to determine the sorting destinations of files

        do_recurse : boolean
            should directories in the `source_dir` be traversed recursively?

        unknown_suffix : string
            suffix string that is joined with the file extension for
            use when there is no ruleset rule for it.

        no_process: set()
            names of toplevel directories that should not be traversed

        directories_dest : string
            `dest_dir` relative or absolute path  where all directories must be moved

        other_files : string
           `dest_dir` relative or absolute directory path where all files of unknown type are relocated to.

        do_remove_empty_dirs : boolean
            toggles recursive empty directory removal

        dest_dir: string
            the directory to organize your files into, instead of doing it in-place,
            if not set, it will be set to `source_dir`

        do_sort_directories: boolean
            should directories also be considered for sorting?
        """

        directories_dest = directories_dest or "directories"
        other_files = other_files or "other"
        self.unknown_file_suffix = unknown_suffix or "_files"

        dest_dir = dest_dir or source_dir

        if os.path.exists(source_dir) and not os.path.isdir(source_dir):
            raise OSError("Directory to organize does not exist or is a file: {}".format(source_dir))

        if not os.path.isdir(dest_dir):
            log.warn("Destination directory does not exist, creating: %s", dest_dir)
            fs.make_path(dest_dir)

        self.path_source = os.path.abspath(source_dir)
        self.path_dest = os.path.abspath(dest_dir)

        self.sort_rule = sort_rule

        self.unknown_types = set()

        # directories or files that should not be processed paths are relative
        # to the directory to be sorted and should exclude any starting ./

        self.no_process = no_process or set()

        self.do_remove_empty_dirs = do_remove_empty_dirs
        self.do_recurse = do_recurse
        self.do_process_dirs = do_process_dirs

        if not os.path.isabs(directories_dest):
            directories_dest = os.path.join(self.path_dest, directories_dest)

        if not os.path.isabs(other_files):
            other_files = os.path.join(self.path_dest, other_files)

        self.path_dirs_dest = fs.cjoin(directories_dest, is_dir=True)
        self.path_other_files = fs.cjoin(other_files, is_dir=True)

        # if these folders are within our organize destination don't try and organize them
        if fs.is_subdir(directories_dest, self.path_dest):
            relative = fs.relative_to(directories_dest, self.path_dest)
            self.no_process.add(fs.cjoin(relative))

        if fs.is_subdir(other_files, self.path_dest):
            relative = fs.relative_to(other_files, self.path_dest)
            self.no_process.add(fs.cjoin(relative))

        fs.make_path(self.path_dirs_dest)
        fs.make_path(self.path_other_files)

    def unknown_ext_dirname(self, extension):
        """Returns the destination directory name for a file with unknown extension"""
        return extension.lower() + self.unknown_file_suffix

    @fs.save_cwd
    def organize(self):
        os.chdir(self.path_source)

        def filter_no_process(base, alist):
            """
            Removes items from a list returned by os.walk, in place.
            """
            L = len(alist)
            for idx, i in enumerate(list(reversed(alist))):
                i = fs.cjoin(base, i)
                if i in self.no_process:
                    del alist[L - idx - 1]

        for base, dirs, files in os.walk('.'):
            filter_no_process(base, files)
            filter_no_process(base, dirs)

            for file in files:
                self.process(fs.cjoin(base, file))

            if self.do_process_dirs:
                for dir in dirs:
                    self.process(fs.cjoin(base, dir, is_dir=True))

            if not self.do_recurse:
                del dirs[:]

        if self.do_remove_empty_dirs:
            fs.remove_empty_dirs(self.path_source)

    def process(self, path):
        name = fs.name(path)

        try:
            raw_dst = self.sort_rule.destination(path)

            # a relative destination
            if not os.path.isabs(raw_dst):
                raw_dst = os.path.join(self.path_dest, raw_dst)

            if raw_dst.endswith('/'):
                # the destination is a directory
                dst = fs.cjoin(raw_dst, name)
            else:
                # the destination is absolute, file must go there
                dst = fs.cjoin(raw_dst)

        except UnhandledPathException:
            if fs.is_file(path):
                name = os.path.basename(path)
                ext = fs.extension(name)
                if not ext:
                    dst = fs.cjoin(self.path_other_files, name)
                else:
                    dst = fs.cjoin(self.path_other_files, self.unknown_ext_dirname(ext[1:]), name)
                    self.unknown_types.add(ext[1:])
            elif self.do_recurse:
                # don't relocate a directory while busy recurring
                return
            else:
                dst = fs.cjoin(self.path_dirs_dest, name)

        if os.path.exists(dst):
            log.info("destination exists: `%s` --> `%s`", path, dst)
            return

        fs.make_path(os.path.dirname(dst))
        log.info("move `%s` --> `%s`", path, dst)

        if fs.is_file(path):
            fs.move_file(path, dst)
        else:
            fs.move_dir(path, dst)



from __future__ import unicode_literals
import os



def is_string(obj):
    import sys

    PY3 = sys.version_info[0] == 3

    if PY3:
        string_types = str,
    else:
        string_types = basestring,
    return isinstance(obj, string_types)


def initialize_dir(d, filetypes, paths_to_make):
    """

    Parameters
    ----------
    filetypes: dict
        mapping of strings to strings for filetypes

    paths_to_make: list
        list of string tuples (path, type)
        directories end in / and files do not end in /

    """
    # --- create a mapping file
    d.write('filetypes.py', bytearray(mkfiletypesstr(filetypes), encoding="utf-8"))

    # --- create temp files
    for path in paths_to_make:
        if os.path.isabs(path):
            raise ValueError("Paths must be relative: {}".format(path))

        dirpath, filename = os.path.split(path)

        if dirpath and not os.path.isdir(dirpath):
            d.makedir(dirpath)

        if filename:
            d.write(path, b'')


def mkfiletypesstr(mapping):
    """
    Takes a string to string mapping and
    writes them as a mapping, returns a string
    """
    import inspect
    import textwrap
    from io import StringIO
    s = StringIO()

    # --- first dump any functions definitions
    for v in mapping.values():
        if callable(v):
            source = textwrap.dedent(inspect.getsource(v))
            # accomodate for python3
            try:
                source = unicode(source, encoding="utf-8")
            except NameError:
                pass
            s.write(source)
            s.write('\n')

    s.write('RULES = [')
    for k in mapping:
        value = repr(mapping[k])
        if callable(v):
            value = v.__name__
        s.write('({}, {}),\n'.format(repr(k), value))
    s.write(']')
    s = s.getvalue()
    # print s
    return s


def _build_path_tree(paths, path_set, prefix=None):
    # print paths, prefix
    for path in paths:
        if is_string(path):
            if prefix is not None:
                path = os.path.join(prefix, path)
            dirpath = os.path.dirname(path)

            parts = dirpath.split('/')

            if len(parts) > 1:
                total = parts[0] + '/'
                if not prefix:
                    path_set.add(total)
                for part in parts[1:]:
                    total = total + part + '/'
                    path_set.add(total)
            path_set.add(path)
        else:
            new_prefix = path[0]
            if not path[0].endswith('/'):
                new_prefix += '/'

            if prefix:
                new_prefix = os.path.join(prefix, new_prefix)

            path_set.add(new_prefix)

            _build_path_tree(path[1], path_set, prefix=new_prefix)


def build_path_tree(paths, prefix=None):
    """
    Creates and returns a sorted list of all possible paths that
    can be generateed from paths.

    Examples
    --------
    >>> build_path_tree(['1/2/file'])
    ['1/', '1/2/', '1/2/file']
    >>> build_path_tree(['file', ('d1', ['f1', 'f2']), ('d2', [('d3', ['f1']), 'f1'])])
    ['d1/', 'd1/f1', 'd1/f2', 'd2/', 'd2/d3/', 'd2/d3/f1', 'd2/f1', 'file']
    >>> build_path_tree(['file', ('d1', ['f1', 'f2']), ('d2', [('d3', ['f1']), 'f1'])], 'sorted')
    ['sorted/d1/', 'sorted/d1/f1', 'sorted/d1/f2', 'sorted/d2/', 'sorted/d2/d3/', 'sorted/d2/d3/f1', 'sorted/d2/f1', 'sorted/file']


    Parameters
    ----------
    paths: list
        A list of strings or tuples representing the paths to be created.

    prefix: str
        string to prefix to all created paths

    Returns
    -------

    """
    pathset = set()
    _build_path_tree(paths, pathset, prefix=prefix)
    return sorted(pathset)


def tempdir(function):
    def wrapped(*args, **kwargs):
        from testfixtures import TempDirectory
        with TempDirectory() as d:
            return function(d, *args, **kwargs)

    wrapped.__name__ = function.__name__
    wrapped.__module__ = function.__module__
    return wrapped



from __future__ import print_function

import logging
import re

from ..core.util import is_string

log = logging.getLogger(__name__)

from .basesortrule import BaseSortRule, UnhandledPathException


class DefaultSortRule(BaseSortRule):
    """
    Default sorint implementation, uses the regex definitions as given in the filetypes.py specification file
    """

    def __init__(self, rules):
        super(DefaultSortRule, self).__init__()
        self.rules = rules

    def first_match(self, finditer):
        for match in finditer:
            return match
        return None

    def destination(self, path):
        for R, function in self.rules:
            match = self.first_match(R.finditer(path))
            if match:
                return function(match, path)
        raise UnhandledPathException()

    @classmethod
    def load_from(cls, filepath):
        """
        Loads sorting rules from a text file

        Parameters
        ----------
        filepath

        Returns
        -------
        a DefaultSortRule with all the sorting entries

        """
        namespace = {'__builtins__': __builtins__}
        with open(filepath, 'r') as f:
            exec(f.read(), namespace)

        if 'RULES' not in namespace:
            msg = "Configuration file missing RULES: {}".format(filepath)
            raise RuntimeError(msg)

        rules = []
        for item in namespace['RULES']:
            item = list(item)
            item[0] = re.compile(item[0])
            if callable(item[1]):
                rules.append(tuple(item))
            elif is_string(item[1]):
                # XXX remove after debugging
                item[1] = make_regex_rule_function(item[0], item[1])
                rules.append(item)
            else:
                msg = "Unhandled type in rule list. " \
                      "Second item in pair must be " \
                      "callable or string: " + repr(item[1])
                raise RuntimeError(msg)

        return cls(rules)


def make_regex_rule_function(pattern, dstfmt):
    """
    Creates a new path processing function.
    Does not support moving directories
    """

    def process(match, path):
        # FIXME only match against the filename
        try:
            pargs = [match.group(0)] + list(match.groups())
            destination = dstfmt.format(*pargs,
                                        **match.groupdict())
        except IndexError:
            raise ValueError("Destination string placeholders out of range: {}".format(dstfmt))
        except KeyError as e:
            msg = "Destination string placeholder " \
                  "unknown key {}: {}".format(repr(e.args[0]), dstfmt)
            raise ValueError(msg)
        log.debug("process: RE[%s](%s) --> %s", pattern.pattern, path, destination)
        return destination

    return process



from __future__ import print_function

import logging
import os
import shutil

log = logging.getLogger(__name__)


# --------------------------------------------------------------------------
#  Canonical path functions
# --------------------------------------------------------------------------
def cjoin(path, *paths, **kwargs):
    """
    Canonical join.
    Ensures the following invariants hold:
        - all paths relative to ./ are given without the ./ specifier,
          unless they refer to `.`, in which case ./ would be returned
        - all direcory paths end in '/'

    Examples
    >>> cjoin("foo", "bar", "baz")
    foo/bar/baz
    >>> cjoin("foo", "bar", "baz", is_dir=True)
    foo/bar/baz/
    >>> cjoin('./foo')
    foo
    >>> cjoin('./foo', is_dir=True)
    foo/
    >>> cjoin('.', is_dir=True)
    ./
    >>> cjoin('./')
    ./


    Parameters
    ----------
    path
    paths
    kwargs:
        is_dir: boolean
            is this path a directory?
    Returns
    -------

    """
    joined = os.path.join(path, *paths)
    if kwargs.get('is_dir', False) and not path.endswith('/'):
        joined += '/'

    if joined != './' and joined.startswith('./'):
        joined = joined[2:]

    return joined


def is_file(path):
    """
    tests whether a canonical path refers to a file or a directory
    """
    return not path.endswith('/')


def extension(path):
    return os.path.splitext(path)[1]


def name(path):
    """
    Return the name of the last element in the path. if the path ends in a slash,
    then the name of the directory before the slash is returned
    """
    if path.endswith('/'):
        path = path[:-1]
    return os.path.basename(path)


# --------------------------------------------------------------------------
#  Common
# --------------------------------------------------------------------------
def save_cwd(function):
    def func(*args, **kwargs):
        _pwd = os.getcwd()
        try:
            return function(*args, **kwargs)
        finally:
            os.chdir(_pwd)

    return func


# --------------------------------------------------------------------------
#  File related
# --------------------------------------------------------------------------

def move_file(src, dst):
    if not os.path.isfile(src):
        raise OSError("Source path is not a file: {}".format(src))
    #print "mv  {} --> {}".format(src, dst)
    shutil.move(src, dst)


# --------------------------------------------------------------------------
#  Directory related
# --------------------------------------------------------------------------

def move_dir(src, dst):
    """Moves the source directory INTO the destination"""
    if not os.path.isdir(src):
        raise OSError("Source path is not a directory: {}".format(src))
    #print "mvd {} --> {}".format(src, dst)
    shutil.move(src, dst)

def remove_empty_dirs(path):
    """Recursively removes empty direcotries contained within path"""
    for path, dirs, files in os.walk(path, topdown=False):
        log.debug("[clean_empty.at] %s", path)
        if len(os.listdir(path)) == 0:
            os.rmdir(path)
            log.debug("rmdir %s", path)


def make_path(path):
    """Creates intermediary directories so that the path exists"""
    path = os.path.abspath(path)
    if os.path.exists(path) and not os.path.isdir(path):
        raise OSError("File {} exists, but is not a directory".format(path))
    log.debug("make_path: %s", path)
    if not os.path.exists(path):
        os.makedirs(path)


def _os_path_split_asunder(path, debug=False):
    """
    http://stackoverflow.com/a/4580931/171094
    """
    parts = []
    while True:
        newpath, tail = os.path.split(path)
        if debug: print(repr(path), (newpath, tail))
        if newpath == path:
            assert not tail
            if path: parts.append(path)
            break
        parts.append(tail)
        path = newpath
    parts.reverse()
    return parts


def _get_normalized_parts(path):
    """Returns the path parts in normalized form"""
    return _os_path_split_asunder(os.path.realpath(os.path.abspath(os.path.normpath(path))))


def is_subdir(potential_subdirectory, expected_parent_directory):
    """
    http://stackoverflow.com/a/17624617

    Is the first argument a sub-directory of the second argument?

    :param potential_subdirectory:
    :param expected_parent_directory:
    :return: True if the potential_subdirectory is a child of the expected parent directory

    >>> is_subdir('/var/test2', '/var/test')
    False
    >>> is_subdir('/var/test', '/var/test2')
    False
    >>> is_subdir('var/test2', 'var/test')
    False
    >>> is_subdir('var/test', 'var/test2')
    False
    >>> is_subdir('/var/test/sub', '/var/test')
    True
    >>> is_subdir('/var/test', '/var/test/sub')
    False
    >>> is_subdir('var/test/sub', 'var/test')
    True
    >>> is_subdir('var/test', 'var/test')
    True
    >>> is_subdir('var/test', 'var/test/fake_sub/..')
    True
    >>> is_subdir('var/test/sub/sub2/sub3/../..', 'var/test')
    True
    >>> is_subdir('var/test/sub', 'var/test/fake_sub/..')
    True
    >>> is_subdir('var/test', 'var/test/sub')
    False
    >>> is_subdir(r'C:\\foo\\baz', 'C:\\foo\\bar')
    False
    >>> is_subdir(r'C:\\foo\\bar', 'C:\\foo\\bar')
    False
    """
    # make absolute and handle symbolic links, split into components
    sub_parts = _get_normalized_parts(potential_subdirectory)
    parent_parts = _get_normalized_parts(expected_parent_directory)

    if len(parent_parts) > len(sub_parts):
        # a parent directory never has more path segments than its child
        return False

        # we expect the zip to end with the short path, which we know to be the parent
    return all(part1 == part2 for part1, part2 in zip(sub_parts, parent_parts))


def relative_to(subdir, parentdir):
    """Returns a string that's path relative to a parent directory parent

    >>> relative_to('/foo/bar/', '/foo/')
    'bar'
    >>> relative_to('/foo/bar', '/foo/')
    'bar'
    >>> relative_to('/foo/bar/baz/', '/foo/')
    'bar/baz'
    """
    assert is_subdir(subdir, parentdir)

    sub_parts = _get_normalized_parts(subdir)
    parent_parts = _get_normalized_parts(parentdir)

    split_at = len(parent_parts)
    for i in range(len(parent_parts)):
        if sub_parts[i] != parent_parts[i]:
            split_at = i
            break

    return os.path.sep.join(sub_parts[split_at:])



from __future__ import print_function
import os

import pytest

from . import helper
from testfixtures import TempDirectory, tempdir

from ..core import pysorter


def test_noargs():
    """
    test the script when no other arguments than the directory to sort and
    the filetypes.py file are specified

    >>> assert False
    """
    with TempDirectory() as d:
        os.chdir(d.path)
        filetypes = {
            '.*\\.pdf$': 'docs/'
        }

        to_sort = 'source/'

        to_make = ['file.pdf']

        helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

        args = [to_sort, '--filetypes', 'filetypes.py']
        pysorter.main(args)

        # --- compare sorted
        expected = ['docs/',
                    'docs/file.pdf',
                    'other/',
                    'directories/']

        d.compare(expected=expected, path=to_sort)


def test_captures_1():
    with TempDirectory() as d:
        os.chdir(d.path)
        filetypes = {
            '.*\\.pdf$': 'docs/',
            '([^_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
        }

        to_sort = 'source/'

        to_make = ['file.pdf',
                   'awesome_song.mp3']

        helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

        args = [to_sort, '--filetypes', 'filetypes.py']
        pysorter.main(args)

        # --- compare sorted
        expected = ['docs/',
                    'docs/file.pdf',
                    'music/',
                    'music/awesome/',
                    'music/awesome/song.mp3',
                    'other/',
                    'directories/']
        d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_non_existing_directory(d):
    os.chdir(d.path)
    args = ['tosort']
    pysorter.main(args)


@helper.tempdir
def test_file_as_directory(d):
    with pytest.raises(OSError):
        os.chdir(d.path)
        d.write('file', b'')
        args = ['file']
        pysorter.main(args)


@helper.tempdir
def test_recursive_sort(d):
    os.chdir(d.path)
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '--filetypes', 'filetypes.py']
    pysorter.main(args)

    # --- compare sorted
    expected = ['foo/',
                'nested/',
                'nested/even/',
                'nested/even/deeper/',
                'docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3',
                'other/',
                'directories/']
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_recursive_sort_with_directory_processing(d):
    os.chdir(d.path)
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '--filetypes', 'filetypes.py', '--process-dirs']
    pysorter.main(args)

    # --- compare sorted
    expected = ['foo/',
                'nested/',
                'nested/even/',
                'nested/even/deeper/',
                'docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3',
                'other/',
                'directories/']
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_clean_empty(d):
    os.chdir(d.path)
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^/_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'source/'

    to_make = ['nested/even/deeper/file.pdf',
               'awesome_song.mp3',
               'foo/another_song.mp3', ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '-c', '--filetypes', 'filetypes.py']
    pysorter.main(args)

    # --- compare sorted
    expected = ['docs/',
                'docs/file.pdf',
                'music/',
                'music/awesome/',
                'music/awesome/song.mp3',
                'music/another/',
                'music/another/song.mp3']
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_other_files(d):
    os.chdir(d.path)
    filetypes = {
        '.*\\.pdf$': 'docs/',
        '([^_]*)_([^_]*)\\.(mp3)$': 'music/{1}/{2}.{3}'
    }

    to_sort = 'files'

    to_make = ['300.mp4', ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py']
    pysorter.main(args)

    # --- compare sorted
    expected = [
        'directories/',
        'other/',
        'other/mp4_files/',
        'other/mp4_files/300.mp4'
    ]
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_duplicate(d):
    os.chdir(d.path)
    filetypes = {
    }

    to_sort = 's'

    to_make = ['s/300.mp4', 's/movie/300.mp4', ]

    helper.initialize_dir(d, filetypes, to_make)

    args = [to_sort, '-r', '--filetypes', 'filetypes.py']
    pysorter.main(args)

    # --- compare sorted
    expected = [
        'directories/',
        'movie/',
        'movie/300.mp4',
        'other/',
        'other/mp4_files/',
        'other/mp4_files/300.mp4'
    ]
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_no_extension(d):
    os.chdir(d.path)
    filetypes = {
    }

    to_sort = 'files/'

    to_make = ['no_extension', ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-r', '--filetypes', 'filetypes.py']
    pysorter.main(args)

    # --- compare sorted
    expected = [
        'directories/',
        'other/',
        'other/no_extension'
    ]
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_no_recurse_dir(d):
    os.chdir(d.path)
    filetypes = {}

    to_sort = 'files/'

    to_make = [('movies',
                ['1.mp4', 'deep/2.mp4', '3.mp4']),
               'emptydir/'
               ]

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes.py', '--process-dirs']
    pysorter.main(args)

    # --- compare sorted
    expected = [
                   'directories/',
                   'other/',
               ] + helper.build_path_tree(to_make, 'directories/')
    d.compare(expected=expected, path=to_sort)

def test_print_version(capsys):
    try:
        pysorter.main(['--version'])
        assert False, 'did not print out version.'
    except SystemExit:
        out, err = capsys.readouterr()
        from .. import __version__ 
        
        # Python 3.4 & 3.5 print out the version on  stdout instead
        # of stderr.
        try:
            assert err.strip() == __version__
        except AssertionError:
            assert out.strip() == __version__


@helper.tempdir
def test_write_unknown_types_correct(d):
    os.chdir(d.path)
    filetypes = {
        r'\.pdf$': 'docs/'
    }

    unknown = 'unknown.txt'
    to_sort = 'files/'

    u_files = ['movie.mp4', 'kerry.mp3', 'phantom.mp3']
    to_make = ['thesis.pdf'] + u_files

    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    # --- compare sorted
    args = [to_sort, '-u', unknown, '--filetypes', 'filetypes.py']
    pysorter.main(args)

    data = set(_ for _ in d.read(unknown, encoding="utf8").split('\n') if _)

    u_exts = (os.path.splitext(_)[1] for _ in u_files if _[1])
    u_exts = set(_[1:] for _ in u_exts)


    assert len(data) == len(u_exts) and not data.difference(u_exts)



from __future__ import print_function
import os

import pytest

from . import helper
from ..core import pysorter


@helper.tempdir
def test_replacement_pattern(d):
    os.chdir(d.path)
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{keyword}/{name}.{ext}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'cruel/',
        'cruel/hello.pdf',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_bad_keyword_replacement_pattern(d):
    os.chdir(d.path)
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{unknown}/{keyword}/{name}.{ext}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    with pytest.raises(ValueError):
        pysorter.main(args)


@helper.tempdir
def test_bad_numbered_replacement_pattern(d):
    os.chdir(d.path)
    filetypes = {
        '(?P<name>\w+)_(?P<keyword>\w+)\.(?P<ext>pdf)$': '{keyword}/{name}.{ext}/{4}'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    with pytest.raises(ValueError):
        pysorter.main(args)


@helper.tempdir
def test_callable_configuration(d):
    def comparison_function(match, entity):
        return 'foobar/'

    os.chdir(d.path)
    filetypes = {
        '\.pdf$': comparison_function
    }

    to_sort = 'files/'
    dest = 'sorted/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '-d', dest]

    pysorter.main(args)
    # --- compare sorted
    expected = [
        'foobar/',
        'foobar/hello_cruel.pdf',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=dest)


@helper.tempdir
def test_rule_behaviour_into_directory_for_file(d):
    os.chdir(d.path)
    filetypes = {
        '\.pdf$': 'pdf/'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/hello_cruel.pdf',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_rule_behaviour_to_for_file(d):
    os.chdir(d.path)
    filetypes = {
        '.*\.pdf$': 'pdf'
    }

    to_sort = 'files/'

    to_make = ['hello_cruel.pdf']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py']

    # --- compare sorted
    expected = [
        'pdf',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_rule_behaviour_into_directory_for_dir(d):
    os.chdir(d.path)
    filetypes = {
        'foo/$': 'pdf/'
    }

    to_sort = 'files/'

    to_make = ['foo/']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs']

    # --- compare sorted
    expected = [
        'pdf/',
        'pdf/foo/',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=to_sort)


@helper.tempdir
def test_rule_behaviour_to_for_dir(d):
    os.chdir(d.path)
    filetypes = {
        'foo/$': 'pdf'
    }

    to_sort = 'files/'

    to_make = ['foo/']
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-t', 'filetypes.py', '--process-dirs']

    # --- compare sorted
    expected = [
        'pdf/',
        'other/',
        'directories/']
    pysorter.main(args)
    d.compare(expected=expected, path=to_sort)


from __future__ import print_function

import logging

log = logging.getLogger(__name__)


class UnhandledPathException(Exception): pass


class BaseSortRule(object):
    """
    Base implementation that all sorting rules must extend.
    """

    def destination(self, path):
        """
        Process the given path. If path ends in a `/` it's a directory else it's a file.
        This method can return an absolute or relative destination.

        When a destination ends in `/` the item at the path will me moved *inside*
        that destination, else the item will be moved to the location *at* destination.

        Raises
        ------
        UnhandledPathException:
            if the method cannot handle the given path

        """
        raise NotImplementedError()



from __future__ import print_function
import os

import pytest

from . import helper
from testfixtures import tempdir

from ..core import pysorter


@helper.tempdir
def test_bad_filetypes(d):
    os.chdir(d.path)
    filetypes = {}

    to_sort = 'files/'

    to_make = []
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '--filetypes', 'filetypes']

    with pytest.raises(OSError):
        pysorter.main(args)


@helper.tempdir
def test_unknown_filetypes_created(d):
    os.chdir(d.path)
    filetypes = {}

    to_sort = 'files/'

    to_make = []
    helper.initialize_dir(d, filetypes, helper.build_path_tree(to_make, to_sort))

    args = [to_sort, '-u', 'unknown_files']

    # --- compare sorted
    expected = helper.build_path_tree(['',
                                       'other/',
                                       'directories/'],
                                      to_sort) + \
               ['unknown_files',
                'filetypes.py']

    pysorter.main(args)
    d.compare(expected=expected)



"""
This file defines patterns and destination directories for files and directories matching the
given pattern.

    - Directory names can be assumed to always end in a `/`
    - For file rules, when the destination ends in  a `/`, the file would be placed
      inside the destination with that name. If the destination did not end in a `/`
      then the file will be renamed to the given destination, for example assume the
      following directory that is to be sorted.

      directory/
        - file1.pdf

    Consider the following rules and corresponding behaviour
        - RULE:      ('.*\.pdf', 'pdfs/')
          BEHAVIOUR: file1.pdf --> pdfs/file1.pdf

        - RULE:      ('.*\.pdf', 'pdfs')
          BEHAVIOUR: file1.pdf --> pdfs

    This applies similarly to directories

"""
RULES = [
    (r'\.a2w$', 'alice_projects/'),
    (r'\.gz$', 'archives/'),
    (r'\.ace$', 'archives/'),
    (r'\.zip$', 'archives/'),
    (r'\.cab$', 'archives/windowscab/'),
    (r'\.tar$', 'archives/'),
    (r'\.rar$', 'archives/'),
    (r'\.7z$', 'archives/'),
    (r'\.bz2$', 'archives/'),
    (r'\.jar$', 'archives/jars/'),
    (r'\.aup$', 'projects/audacity_projects/'),
    (r'\.bwg$', 'projects/brain_wave_generator_projects/'),
    (r'\.sbk$', 'projects/scrapbook_factory/'),
    (r'\.cdr$', 'projects/corel/'),
    (r'\.mp3$', 'audio/'),
    (r'\.amr$', 'audio/amr/'),
    (r'\.m3u$', 'audio/playlist/'),
    (r'\.wav$', 'audio/'),
    (r'\.ogg$', 'audio/'),
    (r'\.midi$', 'audio/midi/'),
    (r'\.mid$', 'audio/midi/'),
    (r'\.wma$', 'audio/'),
    (r'\.flac$', 'audio/'),
    (r'\.m4a$', 'audio/'),
    (r'\.m4r$', 'audio/'),
    (r'\.mp4$', 'videos/'),
    (r'\.wmv$', 'videos/'),
    (r'\.flv$', 'videos/flash/'),
    (r'\.avi$', 'videos/'),
    (r'\.3gp$', 'videos/3gp/'),
    (r'\.swf$', 'flash/swf/'),
    (r'\.m2ts$', 'videos/'),
    (r'\.m1v$', 'videos/'),
    (r'\.mov$', 'videos/'),
    (r'\.mp2$', 'videos/'),
    (r'\.mpe$', 'videos/'),
    (r'\.mpg$', 'videos/'),
    (r'\.mpeg$', 'videos/'),
    (r'\.jpeg$', 'images/jpg/'),
    (r'\.jpg$', 'images/jpg/'),
    (r'\.ico$', 'images/icons/'),
    (r'\.gif$', 'images/gif/'),
    (r'\.bmp$', 'images/bitmap/'),
    (r'\.png$', 'images/png/'),
    (r'\.svg$', 'images/svg/'),
    (r'\.xcf$', 'images/xcf/'),
    (r'\.tif$', 'images/tif/'),
    (r'\.rtf$', 'documents/writing/'),
    (r'\.nfo$', 'documents/info/'),
    (r'\.diz$', 'documents/info/'),
    (r'\.xls$', 'documents/spreadsheets/'),
    (r'\.xlsx$', 'documents/spreadsheets/'),
    (r'\.ods$', 'documents/spreadsheets/'),
    (r'\.tex$', 'documents/latex/'),
    (r'\.odt$', 'documents/writing/'),
    (r'\.doc$', 'documents/writing/'),
    (r'\.docx$', 'documents/writing/'),
    (r'\.wpd$', 'documents/writing/'),
    (r'\.wps$', 'documents/writing/'),
    (r'\.odb$', 'documents/databases/'),
    (r'\.pps$', 'documents/presentations/'),
    (r'\.ppt$', 'documents/presentations/'),
    (r'\.pptx$', 'documents/presentations/'),
    (r'\.txt$', 'documents/plain_text/'),
    (r'\.pdf$', 'documents/pdf/'),
    (r'\.msg$', 'documents/email/outlook/'),
    (r'\.msi$', 'installers/microsoft/'),
    (r'\.deb$', 'installers/debian/'),
    (r'\.exe$', 'installers/microsoft/'),
    (r'\.sis$', 'instalers/symbian/'),
    (r'\.apk$', 'instalers/android/'),
    (r'\.mht$', 'internet/saved_websites/'),
    (r'\.htm$', 'internet/saved_websites/'),
    (r'\.html$', 'internet/saved_websites/'),
    (r'\.url$', 'internet/url/'),
    (r'\.torrent$', 'internet/torrents/'),
    (r'\.xml$', 'xml/'),
    (r'\.vcs$', 'calendar/'),
    (r'\.vol$', 'virtual_encrypted_disk/'),
    (r'\.reg$', 'windowssystem/registry/'),
    (r'\.lnk$', 'windowssystem/shortcut/'),
    (r'\.ini$', 'windowssystem/configuration/ini/'),
    (r'\.inf$', 'windowssystem/configuration/inf/'),
    (r'\.c$', 'source/c/'),
    (r'\.cpp$', 'source/cpp/'),
    (r'\.py$', 'source/python/'),
    (r'\.java$', 'source/java/'),
    (r'\.c#$', 'source/c#/'),
    (r'\.dat$', 'data_files/'),
    (r'\.csv$', 'data_files/'),
    (r'\.iso$', 'disc_images/iso/'),
    (r'\.xpi$', 'firefox/extensions/'),
    (r'\.jad$', 'javame/jads/'),
    (r'\.ttf$', 'fonts/true_type/'),
    (r'\.gmx$', 'games/aoe_saved_games/'),
    (r'\.asb$', 'misc/hymn_assembler/'),
    (r'\.adr$', 'opera/addressbook_backups/'),
    (r'\.p2p$', 'peerguardian/lists/'),
    (r'\.vkp$', 'sony_ericsson/patches/'),
    (r'\.hid$', 'sony_ericsson/hid/'),
    (r'\.aswcs$', 'avast/themes/'),
    (r'\.vcf$', 'contacts/'),
    (r'\.cer$', 'certificates/'),
]



"""
A utility script to translate from the new filetypes.txt format
to the new .py alternative
"""

import re
import sys
import os
import argparse
# Comments

DELIM = '-->'
P_COMMENT = re.compile(r'(?<!\\)#[^\n]*\n?')
P_ITEM = re.compile(r'^([^\s]*)\s*' + re.escape(DELIM) + r'\s*([^\s]*?)\s*$', re.MULTILINE)


def process_arguments():
    parser = argparse.ArgumentParser()

    parser.add_argument('infile', help='the input file to traslate')

    return parser.parse_args()

def main(infile):
    with open(infile, 'r') as f:
        data = f.read()

    data = P_COMMENT.sub('', data)
    print '['
    for m in P_ITEM.finditer(data):
        regex, dest = m.groups()
        print "    ({}, {}),".format(repr(regex), repr(dest))
    print ']'

if __name__ == '__main__':
    args = process_arguments()
    main(args.infile)



from os.path import *


def script_directory():
    """Returns the full path to where this script located"""
    return abspath(join(dirname(abspath(__file__)), '../'))


def write_unknown(unknowns, path):
    """Appends all the unknown filetypes to the file at the given path.
    A filetype per line"""
    with open(path, 'a') as ufile:
        for i in unknowns:
            ufile.write(i)
            ufile.write('\n')


def is_string(obj):
    import sys

    PY3 = sys.version_info[0] == 3

    if PY3:
        string_types = str,
    else:
        string_types = basestring,
    return isinstance(obj, string_types)



import os
import sys

from setuptools import setup, find_packages, Command
from commands import *

tests_require=['pytest-cov', 'pytest', 'testfixtures']

setup(
    name=name,
    version=read_version(),
    description='A regex based file organizer',
    long_description=open(os.path.join(base_dir, 'description.txt')).read().strip(),
    license='GPL',
    url='https://github.com/chriscz/pySorter',

    author='Chris Coetzee',
    author_email='chriscz93@gmail.com',

    packages=find_packages(),
    setup_requires=['pytest-runner'],
    tests_require=tests_require,

    include_package_data=True,
    zip_safe=False,

    cmdclass={
        'test': PyTestCommand,
        'coverage': CoverageCommand,
        'bump': BumpVersionCommand,
    },

    entry_points={
        "console_scripts": ['pysorter=pysorter.core.pysorter:main']
    },
    extras_require=dict(
        build=['twine', 'wheel', 'setuptools-git'],
        test=['pytest', 'testfixtures', 'pytest-cov'],
    ),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
    ]
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""The main implementation for calling pysorter from the commandline"""

from __future__ import print_function

import argparse
import logging
import os
import sys

from . import util
from .sorter import PySorter

from ..rules import DefaultSortRule

log = logging.getLogger(__name__)


def validate_arguments(args):
    """
        Checks whether the paths and options
        provided as commandline arguments are valid.
        The application exits if they are not.
    """
    if args.filetypes:
        args.filetypes = os.path.abspath(args.filetypes)
    else:
        args.filetypes = os.path.join(util.script_directory(), 'filetypes.py')

    if not os.path.isfile(args.filetypes):
        raise OSError("Filetypes is not a file or does not exist: {}".format(args.filetypes))

    if args.unknown_filetypes:
        args.unknown_filetypes = os.path.abspath(args.unknown_filetypes)
    

    return args


def parse_args(args=None):
    """Create an argument parser"""
    from .. import __version__
    parser = argparse.ArgumentParser(description='Reorganizes files and directories in a directory according to certain rules')

    parser.add_argument('directory',
                        help='The directory to be organized')

    parser.add_argument('-d', '--destination',
                        help='The destination directory to move organized files to.',
                        dest='dest_dir',
                        default=None)

    parser.add_argument('-p', '--process-dirs',
                        help='Should directories be included in the processing?',
                        action='store_true',
                        dest='do_process_dirs')

    parser.add_argument('-t', '--filetypes',
                        help='File containing file types [Default: filetypes.py]',
                        default=None)

    parser.add_argument('-m', '--move-dirs-to',
                        help='Move unprocessed directories here [Default: directories/]',
                        dest='directories_dest')

    parser.add_argument('-o', '--other-files',
                        help='Move unprocessed files here [Default: other/]')

    parser.add_argument('-u', '--unhandled-filetypes',
                        help='Write the extensions of unhandled filetypes to this file',
                        dest='unknown_filetypes')

    parser.add_argument('-r', '--recursive',
                        help='Recursively organize directories',
                        action='store_true',
                        dest='do_recurse')

    parser.add_argument('-c', '--remove-empty-dirs',
                        help='Recursively removes all empty directories in the directory being organized.',
                        action='store_true',
                        dest='do_remove_empty_dirs')

    parser.add_argument('-V', '--version',
                        action="version",
                        version=__version__,
                        help='Prints out the current version of pysorter')

    return validate_arguments(parser.parse_args(args))


def main(args=None):
    logging.basicConfig()
    args = parse_args(args)

    rules = DefaultSortRule.load_from(args.filetypes)

    topass = dict(vars(args))

    # --- remove processed arguments
    del topass['directory']
    del topass['filetypes']
    del topass['unknown_filetypes']

    sorter = PySorter(args.directory, rules, **topass)
    sorter.organize()

    # write out all the unknown file types
    if args.unknown_filetypes:
        util.write_unknown(sorter.unknown_types, args.unknown_filetypes)
    return 0

if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    # sys.argv = ['pySorter.py', '/tmp/test/','-t', '/home/chris/Development/soft_dev/pySorter/pysorter/filetypes.txt', '-r', '-c']
    sys.exit(main())
